#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for ArchiveService """

# pylint: disable=C0103, E0401, R0201, W0621
import http
import json
from typing import Dict
from unittest.mock import MagicMock, patch

import pytest

from workspaces.system.services.archive_service import (
    ArchiveService,
    PauseState,
    RunState,
    StandardCalibrationManager,
    StopState,
)

SPL = "uid://evla/execblock/8fbfb54b-d141-42fe-b079-609339a69cba"
ROUTER_PATCH = "workspaces.system.services.archive_service.Router"


@pytest.fixture(scope="module")
def fake_metadata() -> json:
    return json.dumps(
        {
            "spl": SPL,
            "bands": "Ka",
            "array_config": "D",
            "obs_start_time": "1234567",
            "obs_end_time": "123598898",
        }
    )


class TestArchiveService:
    """Test class for ArchiveService"""

    def test_on_ingestion_complete(self, fake_metadata: Dict[str, str]):
        ingestion_message = {
            "user": "test_user",
            "hostname": "hostname01",
            "timestamp": {
                "date": {"year": 2021, "month": 6, "day": 17},
                "time": {"hour": 18, "minute": 24, "second": 52, "nano": 18355000},
            },
            "application": "unknown",
            "request": "snek asks nice-like",
            "version": 1.0,
            "message": "Ingestion complete",
            "logData": {
                "ingestion_type": "rawdata",
                "execblock_id": "29223944",
                "override_db_screen": "False",
                "fileset_id": "13B-014.sb29151475.eb29223944.56810.442529050924",
                "telescope": "EVLA",
                "schedblock_id": "29151475",
                "schedblock_type": "SCIENCE",
                "project_code": "13B-014",
                "file_count": "1",
                "profile": None,
            },
        }

        expected_spl = SPL
        with patch(
            "workspaces.system.services.archive_service.WrestObservationMetadata.wrest_observation_info",
        ) as mock_wrest_obs_metadata:
            mock_wrest_obs_metadata.return_value = fake_metadata
            with patch(ROUTER_PATCH):
                service = ArchiveService()
                service.stdcal = MagicMock()
                service.stdcal.ingestion_complete = MagicMock()
                (
                    spl,
                    execblock_id,
                    fileset_id,
                    project_code,
                ) = service.on_ingestion_complete(**ingestion_message)
                assert service.stdcal.ingestion_complete.call_count == 1
            assert spl == expected_spl
            assert execblock_id == "29223944"
            assert fileset_id == "13B-014.sb29151475.eb29223944.56810.442529050924"
            assert project_code == "13B-014"

    def test_on_ingestion_complete_without_optional_parameters(self, fake_metadata: Dict[str, str]):
        ingestion_message = {
            "logData": {
                "fileset_id": "13B-014.sb29151475.eb29223944.56810.442529050924",
            },
        }

        expected_spl = "uid://evla/execblock/8fbfb54b-d141-42fe-b079-609339a69cba"
        with patch(
            "workspaces.system.services.archive_service.WrestObservationMetadata.wrest_observation_info",
        ) as mock_wrest_obs_metadata:
            mock_wrest_obs_metadata.return_value = fake_metadata
            with patch(ROUTER_PATCH):
                service = ArchiveService()
                service.stdcal = MagicMock()
                service.stdcal.ingestion_complete = MagicMock()
                (
                    spl,
                    execblock_id,
                    fileset_id,
                    project_code,
                ) = service.on_ingestion_complete(**ingestion_message)
                assert service.stdcal.ingestion_complete.call_count == 1
            assert spl == expected_spl
            assert execblock_id == ""
            assert fileset_id == "13B-014.sb29151475.eb29223944.56810.442529050924"
            assert project_code == ""

    def test_on_ingest_complete_returns_nothing_for_vlass(self, fake_metadata: Dict[str, str]):
        """
        Do not submit VLASS calibrations for ingestion.
        """
        ingestion_message = {
            "user": "test_user",
            "hostname": "hostname01",
            "timestamp": {
                "date": {"year": 2021, "month": 6, "day": 17},
                "time": {"hour": 18, "minute": 24, "second": 52, "nano": 18355000},
            },
            "application": "unknown",
            "request": "snek asks nice-like",
            "version": 1.0,
            "message": "Ingestion complete",
            "logData": {
                "ingestion_type": "rawdata",
                "execblock_id": "29223944",
                "override_db_screen": "False",
                "fileset_id": "VLASS1.1.sb29151475.eb29223944.56810.442529050924",
                "telescope": "EVLA",
                "schedblock_id": "29151475",
                "schedblock_type": "SCIENCE",
                "project_code": "VLASS1.1",
                "file_count": "1",
                "profile": None,
            },
        }
        with patch(
            "workspaces.system.services.archive_service.WrestObservationMetadata.wrest_observation_info",
        ) as mock_wrest_obs_metadata:
            mock_wrest_obs_metadata.return_value = fake_metadata
            with patch(ROUTER_PATCH):
                service = ArchiveService()
                service.stdcal = MagicMock()

                with pytest.raises(TypeError) as t_err:
                    service.stdcal.ingestion_complete = MagicMock()
                    () = service.on_ingestion_complete(**ingestion_message)
                    result = service.on_ingestion_complete(**ingestion_message)
                    assert str(t_err) == "TypeError: cannot unpack non-iterable NoneType object"
                    assert result is None
                    assert service.stdcal.ingestion_complete.call_count == 1

    def test_on_ingest_complete_returns_nothing_for_tsky(self, fake_metadata: Dict[str, str]):
        """
        Do not submit TSKY calibrations for ingestion.
        """
        ingestion_message = {
            "user": "test_user",
            "hostname": "hostname01",
            "timestamp": {
                "date": {"year": 2021, "month": 6, "day": 17},
                "time": {"hour": 18, "minute": 24, "second": 52, "nano": 18355000},
            },
            "application": "unknown",
            "request": "snek asks nice-like",
            "version": 1.0,
            "message": "Ingestion complete",
            "logData": {
                "ingestion_type": "rawdata",
                "execblock_id": "29223944",
                "override_db_screen": "False",
                "fileset_id": "TSKY0001.sb29151475.eb29223944.56810.442529050924",
                "telescope": "EVLA",
                "schedblock_id": "29151475",
                "schedblock_type": "SCIENCE",
                "project_code": "TSKY0001",
                "file_count": "1",
                "profile": None,
            },
        }
        with patch(
            "workspaces.system.services.archive_service.WrestObservationMetadata.wrest_observation_info",
        ) as mock_wrest_obs_metadata:
            mock_wrest_obs_metadata.return_value = fake_metadata
            with patch(ROUTER_PATCH):
                service = ArchiveService()
                service.stdcal = MagicMock()

                with pytest.raises(TypeError) as t_err:
                    service.stdcal.ingestion_complete = MagicMock()
                    () = service.on_ingestion_complete(**ingestion_message)
                    result = service.on_ingestion_complete(**ingestion_message)
                    assert str(t_err) == "TypeError: cannot unpack non-iterable NoneType object"
                    assert result is None
                    assert service.stdcal.ingestion_complete.call_count == 1

    def test_create_std_calibration_request(self, fake_metadata: Dict[str, str]):
        spl = SPL
        params = {
            "parameters": {
                "product_locator": spl,
                "user_email": "workspaces-analysts-test@listmgr.nrao.edu",
                "metadata": fake_metadata,
            }
        }

        with patch(ROUTER_PATCH):
            with patch(
                "workspaces.system.services.archive_service.requests.post",
            ) as mock_post:
                mock_post.return_value = MagicMock()
                mock_post.return_value.status_code = http.HTTPStatus.OK
                StandardCalibrationManager().create_stdcal_request(spl, fake_metadata)
                mock_post.assert_called_with(
                    "http://capability:3457/capability/std_calibration/request/create", json=params
                )


class TestStandardCalibrationState:
    def test_dispatch(self):
        with patch("workspaces.system.services.archive_service.StandardCalibrationManager.state_name", "RUN"):
            with patch("workspaces.system.services.archive_service.RunState.ingest_complete") as ic:
                stdcal = StandardCalibrationManager()
                stdcal.ingestion_complete("fakespl", {})
                assert ic.call_count == 1

        with patch("workspaces.system.services.archive_service.StandardCalibrationManager.state_name", "PAUSE"):
            with patch("workspaces.system.services.archive_service.PauseState.ingest_complete") as ic:
                stdcal = StandardCalibrationManager()
                stdcal.ingestion_complete("fakespl", {})
                assert ic.call_count == 1

        with patch("workspaces.system.services.archive_service.StandardCalibrationManager.state_name", "STOP"):
            with patch("workspaces.system.services.archive_service.StopState.ingest_complete") as ic:
                stdcal = StandardCalibrationManager()
                stdcal.ingestion_complete("fakespl", {})
                assert ic.call_count == 1

    @staticmethod
    def generate_fake_stdcal():
        fake_stdcal = MagicMock()
        fake_stdcal.create_stdcal_request = MagicMock()
        fake_stdcal.create_stdcal_request.return_value = -2
        fake_stdcal.submit_stdcal_request = MagicMock()
        return fake_stdcal

    def test_run(self):
        fake_stdcal = self.generate_fake_stdcal()
        run = RunState(fake_stdcal)
        run.ingest_complete("fake", {})
        assert fake_stdcal.create_stdcal_request.call_count == 1
        assert fake_stdcal.submit_stdcal_request.call_count == 1

    def test_pause(self):
        fake_stdcal = self.generate_fake_stdcal()
        run = PauseState(fake_stdcal)
        run.ingest_complete("fake", {})
        assert fake_stdcal.create_stdcal_request.call_count == 1
        assert fake_stdcal.submit_stdcal_request.call_count == 0

    def test_stop(self):
        fake_stdcal = self.generate_fake_stdcal()
        run = StopState(fake_stdcal)
        run.ingest_complete("fake", {})
        assert fake_stdcal.create_stdcal_request.call_count == 0
        assert fake_stdcal.submit_stdcal_request.call_count == 0
